require 'rack/state'
require 'rack/mock'

module StateMiddlewareHelpers #:nodoc:
  def valid_token
    lambda { |obj|
      name = obj.split('=')[0]
      obj.match /^#{name}=[^;]{22,}$/
    }
  end

  def expired_token
    lambda { |obj|
      name = obj.split('=')[0]
      obj.match /^#{name}=; max-age=0; expires=Thu, 01 Jan 1970/
    }
  end

  def key
    'token' # default key name
  end

  def get_state(key, token)
    @req.get("/get_state?key=#{key}", {'HTTP_COOKIE' => token})['Set-Cookie']
  end

  def set_state(key, token = nil)
    cookie = token ? {'HTTP_COOKIE' => token} : {}
    @req.get("/set_state?key=#{key}", cookie)['Set-Cookie']
  end

  def del_state(key, token)
    @req.get("/del_state?key=#{key}", {'HTTP_COOKIE' => token})['Set-Cookie']
  end

  def mock_app
    Proc.new do |env|
      req = Rack::Request.new(env)
      key = req.params['key'] || self.key
      skey = 'rack.state.' + key
      data = nil

      case req.path
      when '/get_state'
        data = env[skey].get
      when '/set_state'
        data = env[skey].set(key + 'data')
      when '/del_state'
        env[skey].delete
      end

      Rack::Response.new(data.to_s).to_a
    end
  end
end

module StateMiddlewarePathHelpers #:nodoc:
  def allow(path)
    should.not.raise(SecurityError) { @req.get(path) }
  end

  def disallow(path)
    should.raise(SecurityError) { @req.get(path) }
  end

  def mock_app
    Proc.new do |env|
      if env['rack.state.token'].nil?
        raise SecurityError, 'state unavailable at path'
      else
        Rack::Response.new.finish
      end
    end
  end
end

describe Rack::State do

  describe 'Default Options' do
    extend StateMiddlewareHelpers

    before do
      @req = Rack::MockRequest.new(Rack::State.new(mock_app))
    end

    should 'not return token if state object not accessed' do
      @req.get('/')['Set-Cookie'].should.be.nil
    end

    should 'return token when setting state' do
      set_state(key).should.be.a valid_token
    end

    should 'return previous token' do
      token1 = set_state(key)
      token2 = get_state(key, token1)
      token1.should.equal token2
    end

    should 'not return token after deleting state' do
      token1 = set_state(key)
      token2 = del_state(key, token1)
      token3 = get_state(key, token1)
      token1.should.be.a valid_token
      token2.should.be.an expired_token
      token3.should.be.nil
    end

    should 'return no token when getting state with nonexistent token' do
      get_state(key, "#{key}=nonexistent").should.be.nil
    end

    should 'return new token when setting state with nonexistent token' do
      token = set_state(key, "#{key}=nonexistent")
      token.should.not.equal "#{key}=nonexistent"
      token.should.be.a valid_token
    end

    should 'expire token when deleting state with nonexistent token' do
      token = del_state(key, "#{key}=nonexistent")
      token.should.not.equal "#{key}=nonexistent"
      token.should.be.an expired_token
    end
  end

  describe 'Two Instances Each With Different Stores' do
    require 'tmpdir'
    extend StateMiddlewareHelpers

    before do
      fstore = Rack::State::Store::File.new(@dir = Dir.mktmpdir)
      mstore = Rack::State::Store::Memory.new
      app = Rack::State.new(mock_app, key:'f', store:fstore)
      app = Rack::State.new(app,      key:'m', store:mstore)
      @req = Rack::MockRequest.new(app)
    end

    after do
      FileUtils.remove_dir @dir
    end

    should 'return persistent data via response body on subsequent requests' do
      fres = @req.get('/set_state?key=f')
      mres = @req.get('/set_state?key=m')
      fres.body.should.equal 'fdata'
      mres.body.should.equal 'mdata'
      fres = @req.get('/get_state?key=f', 'HTTP_COOKIE' => fres['Set-Cookie'])
      mres = @req.get('/get_state?key=m', 'HTTP_COOKIE' => mres['Set-Cookie'])
      fres.body.should.equal 'fdata'
      mres.body.should.equal 'mdata'
      fres = @req.get('/del_state?key=f', 'HTTP_COOKIE' => fres['Set-Cookie'])
      mres = @req.get('/del_state?key=m', 'HTTP_COOKIE' => mres['Set-Cookie'])
      fres.body.should.be.empty
      mres.body.should.be.empty
    end

    should 'return tokens when setting state for two different keys' do
      ftoken = set_state('f')
      mtoken = set_state('m')
      ftoken.should.be.a valid_token
      mtoken.should.be.a valid_token
      ftoken.should.not.equal mtoken
    end

    should 'set states and delete one at a time verifing no token returned' do
      ftoken = set_state('f')
      mtoken = set_state('m')
      del_state('f', ftoken).should.be.an expired_token
      get_state('f', ftoken).should.be.nil
      get_state('m', mtoken).should.be.a valid_token
      del_state('m', mtoken).should.be.an expired_token
      get_state('m', mtoken).should.be.nil
    end

    should 'return previous token for each state' do
      ft1 = set_state('f')
      mt1 = set_state('m')
      ft2 = get_state('f', ft1)
      mt2 = get_state('m', mt1)
      ft1.should.equal ft2
      mt1.should.equal mt2
      ft2.should.not.equal mt2
    end

    should 'return new tokens when setting states with nonexistent tokens' do
      ftoken= 'f=nonexistent'
      mtoken= 'm=nonexistent'
      ft = set_state('f', ftoken)
      mt = set_state('m', mtoken)
      ft.should.not.equal ftoken
      ft.should.be.a valid_token
      mt.should.not.equal mtoken
      mt.should.be.a valid_token
      mt.should.not.equal ft
    end
  end

  describe 'Path Restrictions' do
    extend StateMiddlewarePathHelpers

    should 'allow state access for any path by default' do
      @req = Rack::MockRequest.new(Rack::State.new(mock_app))
      allow '/'
      allow '/a'
      allow '/a/b'
    end

    should 'allow state access for specified path only' do
      @req = Rack::MockRequest.new(Rack::State.new(mock_app, path:'/a'))
      allow '/a'
      allow '/a/b'
      disallow '/'
      disallow '/b'
    end

    should 'allow state access for specified multi-level path only' do
      @req = Rack::MockRequest.new(Rack::State.new(mock_app, path:'/a/b'))
      allow '/a/b'
      allow '/a/b/c'
      disallow '/'
      disallow '/a'
      disallow '/b'
    end
  end
end
